# the main file of the program

import sys
import os
from .SpudLoader import SpudLoader
from .Rule import RuleBox

def main():
    if len(sys.argv) != 2:
        raise Exception("Expected a file to process")
    if os.path.isabs(sys.argv[1]):
        initial_path = sys.argv[1]
    else:
        initial_path = os.path.join(os.getcwd(),sys.argv[1])
    loader = SpudLoader()
    loader.recursive_load_parse(initial_path)
    rulebox = RuleBox()
    rulebox.add(loader.rules)
    html = rulebox.resolve_to_html_document(loader.ingredients)
    with open("ResultsTree.html","w") as file:
        file.write(html)

if __name__ == "__main__":
    main()


# defines the Rule class
# a container for Patterns and RuleOutputs
# defines the HoldRule class
# to allow for a Rule which does nothing
# except to signal the end of evaluation
# defines RuleBox to contain instances
# of Rule

from .Ingredient import Ingredient
from .RuleOutput import RuleOutput, NoneRuleOutputInstance
from .Pattern import Pattern
from .ResultsTree import ResultsTree

from copy import copy

class Rule(object):
    def __init__(self,pattern,outputs):
        if not isinstance(pattern,Pattern):
            raise TypeError("Expected Pattern")
        if type(outputs) != list:
            raise TypeError("Expected list")
        if len(outputs) == 0:
            raise ValueError("Expected Nonempty list")
        if not all(map(lambda x: isinstance(x,RuleOutput),outputs)):
            raise TypeError("Expected list of RuleOutput")
        self.pattern = pattern
        self.outputs = outputs
    def matches(self,ig):
        # returns match_token
        return self.pattern.matches(ig)
    def specificity(self):
        return self.pattern.specificity
    def apply(self,ig,match_token):
        # match_token is that which was returned from
        # self.matches
        return list(map(lambda x: x.apply(ig,match_token),self.outputs))
    def priority(self):
        # self.outputs has nonzero length
        return self.outputs[0].priority()

class HoldRule(Rule):
    def __init__(self,pattern,priority):
        super().__init__(pattern,[NoneRuleOutputInstance])
        if type(priority) != int:
            raise TypeError("Expected int")
        self._priority = priority
    def priority(self):
        return self._priority

class RuleBox(object):
    def __init__(self):
        self.rules = []
    def add(self,rule):
        if isinstance(rule,Rule):
            self.rules.append(rule)
        else:
            if type(rule) == list:
                for elem in rule:
                    self.add(elem)
            else:
                raise TypeError("Expected Rule")
    def _search(self,ig,mask):
        # if is Ingredient
        # mask is dict(int->str)
        # the keys indicate which indicies to not check
        best = None
        best_index = None
        best_match_token = None
        for i in range(len(self.rules)):
            if i in mask:
                continue
            rule = self.rules[i]
            if best is not None:
                if rule.priority() > best.priority():
                    continue
                elif rule.priority() == best.priority():
                    if rule.specificity() <= best.specificity():
                        continue
            # we are good to test
            match_token = rule.matches(ig)
            if (match_token[0] != 0):
                best = rule
                best_index = i
                best_match_token = match_token
        return best, best_index, best_match_token
    def resolve(self,output_array,ig,mask=None):
        if mask is None:
            mask = dict()
        if type(output_array) != list:
            raise TypeError("Expected list")
        if not isinstance(ig,Ingredient):
            raise TypeError("Expected Ingredient")
        while True:
            rule, index, match_token = self._search(ig,mask)
            if (rule is None) or isinstance(rule,HoldRule):
                # we are done
                output_array.append(ig)
                return
            results = rule.apply(ig,match_token)
            if len(results) == 1:
                if results[0] is ig:
                    # we did not change ig
                    # we did not apply a rule
                    mask[index] = "skip"
                else:
                    mask[index] = "applied"
                    ig = results[0]
                    for key in list(filter(lambda key: mask[key] == "skip", mask)):
                        del mask[key]
            else:
                for result in results:
                    submask = copy(mask)
                    if result is ig:
                        # we did not change ig
                        # we did not apply a rule
                        submask[index] = "skip"
                    else:
                        submask[index] = "applied"
                    self.resolve(output_array,result,mask=submask)
                return
    def resolve_to_html_document(self,ig_array):
        if type(ig_array) != list:
            raise TypeError("Expected list")
        if not all(map(lambda x: type(x) == Ingredient, ig_array)):
            raise TypeError("Expected list of Ingredient objects")
        tree = ResultsTree()
        for ig in ig_array:
            output_array = []
            self.resolve(output_array,ig)
            for output_ig in output_array:
                tree.add_ingredient(output_ig.strip())
        return tree.as_html_document()


# defines a RuleOutput class
# and its subclasses
# these perform singular operators on Ingredient objects

from .MyNumber import MyNumber
from .Ingredient import Ingredient
from .FormatCheck import listOfStrings, listOfTupleStringBools

class RuleOutput(object):
    def __init__(self):
        raise Exception("creating an instance of RuleOutput base class is forbidden")
    @staticmethod
    def check_conformity(ig,match_token,match_token_length):
        # check that an input to apply makes sense
        if not isinstance(ig,Ingredient):
            raise TypeError("Expected ig to be an Ingredient")
        if type(match_token) != tuple:
            raise TypeError("Expected tuple")
        if type(match_token_length) != int:
            raise TypeError("Expected int")
        if len(match_token) != match_token_length:
            raise ValueError("Expected tuple of length "+str(match_token_length))
        if not all(map(lambda x: type(x) == int, match_token)):
            raise TypeError("Expected tuple filled with ints")

class NoneRuleOutput(RuleOutput):
    def __init__(self):
        pass
    @staticmethod
    def apply(ig,match_token):
        return ig
    @staticmethod
    def priority():
        return 1

NoneRuleOutputInstance = NoneRuleOutput()

class RenamingRuleOutput(RuleOutput):
    def __init__(self,output_name):
        if output_name is not None:
            listOfStrings(output_name)
        self.output_name = output_name
    def apply(self,ig,match_token):
        if self.output_name is None:
            return ig
        # match token must come from SinglePattern
        self.check_conformity(ig,match_token,2)
        output = ig.duplicate()
        if match_token[0] == 1:
            # editing unit
            output.unit = output.unit[:max(0,match_token[1])] + self.output_name
        else:
            #editing name
            output.name = output.name[:max(0,match_token[1])] + self.output_name
        return output
    @staticmethod
    def priority():
        return 2

class PrefixingRuleOutput(RuleOutput):
    def __init__(self,prefix):
        listOfStrings(prefix)
        self.prefix = prefix
    def apply(self,ig,match_token):
        # match token must come from SinglePattern
        self.check_conformity(ig,match_token,2)
        if match_token[1] > 0:
            # if the input already has a prefix,
            # we shouldn't change it
            return ig
        output = ig.duplicate()
        if match_token[0] == 1:
            # editing unit
            output.unit = self.prefix + output.unit
        else:
            # editing name
            output.name = self.prefix + output.name
        return output
    @staticmethod
    def priority():
        return 3

class InsertingRuleOutput(RuleOutput):
    def __init__(self,pattern_size,insertion):
        if type(pattern_size) != int:
            raise TypeError("Expeected int")
        if insertion != []:
            listOfStrings(insertion)
        self.pattern_size = pattern_size
        self.insertion = insertion
    def apply(self,ig,match_token):
        # match token must come from SinglePattern
        self.check_conformity(ig,match_token,2)
        output = ig.duplicate()
        if match_token[0] == 1:
            # editing unit
            output.unit = output.unit[:max(0,match_token[1])] + self.insertion + output.unit[max(0,match_token[1])+self.pattern_size:]
        else:
            # editing name
            output.name = output.name[:max(0,match_token[1])] + self.insertion + output.name[max(0,match_token[1])+self.pattern_size:]
        return output
    @staticmethod
    def priority():
        return 2

class SingleConvertingRuleOutput(RenamingRuleOutput):
    def __init__(self,ratio,output_name):
        super().__init__(output_name)
        if not isinstance(ratio,MyNumber):
            raise TypeError("Expected MyNumber")
        self.ratio = ratio
    def apply(self,ig,match_token):
        self.check_conformity(ig,match_token,2)
        output = super().apply(ig,match_token).duplicate()
        output.count = output.count * self.ratio
        return output
    @staticmethod
    def priority():
        return 5

class DoubleConvertingRuleOutput(RuleOutput):
    def __init__(self,ratio,output_unit,output_name):
        # ratio, output_unit, and output_name
        # may each be None to indicate a wildcard
        if ratio is not None:
            if not isinstance(ratio,MyNumber):
                raise TypeError("Expected MyNumber")
        if output_unit is not None:
            output_unit = RenamingRuleOutput(output_unit)
        if output_name is not None:
            output_name = RenamingRuleOutput(output_name)
        self.ratio = ratio
        self.output_unit = output_unit
        self.output_name = output_name
    def apply(self,ig,match_token):
        # match_token must be from DoublePattern
        self.check_conformity(ig,match_token,3)
        if self.output_unit is not None:
            ig = self.output_unit.apply(ig,(1,max(0,match_token[1])))
        if self.output_name is not None:
            ig = self.output_name.apply(ig,(2,max(0,match_token[2])))
        if self.ratio is not None:
            ig = ig.duplicate()
            ig.count = ig.count * self.ratio
        return ig
    @staticmethod
    def priority():
        return 5

class PropertiesRuleOutput(RuleOutput):
    def __init__(self,base,edits):
        if not isinstance(base,RuleOutput):
            raise TypeError("Expected child class of RuleOutput")
        listOfTupleStringBools(edits)
        self.base = base
        self.edits = edits
    def apply(self,ig,match_token):
        output = self.base.apply(ig,match_token).duplicate()
        for edit in self.edits:
            # edit is (str,bool)
            if edit[1]:
                output.props.add(edit[0])
            else:
                if edit[0] == "$":
                    # remove all
                    output.props = set()
                elif edit[0] in output.props:
                    output.props.remove(edit[0])
        return output
    def priority(self):
        return self.base.priority()

class DecRuleOutput(RuleOutput):
    def __init__(self):
        pass
    @staticmethod
    def apply(ig,match_token):
        output = ig.duplicate()
        output.count = output.count.as_float()
        return output
    @staticmethod
    def priority():
        return 1

DecRuleOutputInstance = DecRuleOutput()

class FracRuleOutput(RuleOutput):
    def __init__(self):
        pass
    @staticmethod
    def apply(ig,match_token):
        output = ig.duplicate()
        output.count = output.count.as_fraction(10)
        return output
    @staticmethod
    def priority():
        return 1
    
FracRuleOutputInstance = FracRuleOutput()


# defines the SpudLoader class
# to load and parse source files

import os
import hashlib
from copy import copy
from .SourceVersionCheck import source_version_check
import antlr4
from antlr4.error.ErrorListener import ErrorListener
from .Pattern import SinglePattern, DoublePattern
from .Rule import Rule, HoldRule
from .RuleOutput import (
    RenamingRuleOutput, PrefixingRuleOutput, InsertingRuleOutput,
    SingleConvertingRuleOutput, DoubleConvertingRuleOutput,
    DecRuleOutputInstance, FracRuleOutputInstance, PropertiesRuleOutput)
from .SpudLexer import SpudLexer
from .SpudParser import SpudParser

from .MyNumber import MyNumber
from .Ingredient import Ingredient

class MyErrorListener(ErrorListener):
    # from https://stackoverflow.com/questions/32224980/python-2-7-antlr4-make-antlr-throw-exceptions-on-invalid-input
    def __init__(self,filename):
        super(MyErrorListener, self).__init__()
        self.filename = filename
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception("Syntax Error in file: "+str(self.filename)+" on line: "+str(line)+" in column: "+str(column)+" with message: "+str(msg))
    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        pass
    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        pass
    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        pass

class SpudLoader(object):
    def __init__(self):
        self.ingredients = []
        self.rules = []
        self.seen = set() # sha256 hashes of files
    @staticmethod
    def load_file_raw(path):
        if path[-5:] != ".spud":
            path = path + ".spud"
        if os.path.isfile(path):
            with open(path,"rb") as file:
                return file.read()
        else:
            raise Exception("Unable to locate file: "+path)
    def get_tree(self,abs_path,force):
        # abs_path is a string with the absolute file path to load
        # force is a bool. True means load outside the normal import system
        # returns either the syntax tree for the source file
        # or None
        raw = self.load_file_raw(abs_path)
        if not force:
            file_hash = hashlib.sha256(raw).digest()
            if file_hash in self.seen:
                # we have seen this file before
                return None
            else:
                # we have not seen this file before
                # register it so that we do not
                # need to process it again
                self.seen.add(file_hash)
        # check the source version to make sure
        # that it's compatible with this grammar
        try:
            version_compat = source_version_check(raw)
        except:
            raise Exception("Malformed version information in file: "+abs_path)
        if not version_compat[0]:
            raise Exception("Incompatible source version in file: "+abs_path+". Error Message: "+version_compat[1])
        # first convert the input to a string
        # and then to a stream
        raw_string = raw.decode("UTF-8")
        stream = antlr4.InputStream(raw_string)
        # stream contains the input
        lexer = SpudLexer(stream)
        token_stream = antlr4.CommonTokenStream(lexer)
        # token_stream contains a stream of tokens
        # from the input
        parser = SpudParser(token_stream)
        parser.addErrorListener( MyErrorListener(abs_path) )
        tree = parser.spud()
        # tree is the entire syntax tree for the input
        return tree
    @staticmethod
    def extract_imports(output,tree):
        if isinstance(tree,SpudParser.AtimportContext):
            # we found one
            output.append(tree.getText())
        elif isinstance(tree,(SpudParser.SpudContext,
                SpudParser.LineContext,
                SpudParser.LineatContext,
                SpudParser.AtsectionContext)):
            if tree.children is None:
                return
            for child in tree.children:
                SpudLoader.extract_imports(output,child)
    @staticmethod
    def convert_to_absolute_pathnames(base_dir,path):
        if os.path.isabs(path):
            return path
        else:
            output = os.path.join(base_dir,path)
            output = os.path.abspath(output)
            return output
    @staticmethod
    def resolve_import_path(base_dir,statement):
        # we know that @import exists in the statement
        index = statement.find("@import")
        # chop off the @import
        statement = statement[index+7:]
        # chop off all leading spaces
        while True:
            if len(statement) < 1:
                break
            if statement[0] == " " or statement[0] == "\t":
                statement = statement[1:]
            else:
                break
        # chop off all training spaces
        while True:
            if len(statement) < 1:
                break
            if statement[-1] == " " or statement[-1] == "\t":
                statement = statement[:-1]
            else:
                break
        return SpudLoader.convert_to_absolute_pathnames(base_dir,statement)
    def _handle_all_prop_elements(self,tree,context):
        # tree is SpudParser.Propspluselem
        # or SpudParser.Propselem
        # or SpudParser.Propswildelem
        for child in tree.children:
            # all terminal nodes
            text = child.getText()
            if child.getSymbol().type == SpudParser.Namesegment:
                return (text,sign)
            elif text == "$":
                return (text,sign)
            elif text == "+":
                sign = True
            elif text == "-":
                sign = False
    def _handle_all_props(self,tree,context):
        # tree is SpudParser.Propsplus
        # or SpudParser.Props
        # or SpudParser.Propswild
        if tree.children is None:
            return []
        props = []
        for child in tree.children:
            props.append(self._handle_all_prop_elements(child,context))
        return props
    @staticmethod
    def _props_conversion(props_in):
        # props_in is list(tuple(str,bool))
        # we will output the same information
        # in dict(str->bool)
        props_out = {}
        for element in props_in:
            props_out[element[0]] = element[1]
        return props_out
    def _handle_all_names(self,tree,context):
        # tree is SpudParser.Basicname
        # or SpudParser.Powname
        name = []
        if isinstance(tree,SpudParser.PownameContext):
            for child in tree.children:
                if isinstance(child,antlr4.tree.Tree.TerminalNodeImpl):
                    # this can only be !
                    name.append("!")
                else:
                    # this can only be SpudParser.Basicname
                    name += self._handle_all_names(child,context)
        else:
            # all children are terminal nodes
            for child in tree.children:
                if child.getSymbol().type == SpudParser.Namesegment:
                    name.append(child.getText())
        return name
    def _handle_name_and_props(self,tree,context):
        # tree is:
        # SpudParser.Basicnameplusprops
        # SpudParser.Pownamewithprops
        # SpudParser.Wildcardwithprops
        # SpudParser.Basicnamewithpropswild
        # SpudParser.Wildcardwithpropswild
        # SpudParser.Powwildwithprops
        # SpudParser.Basicwildwithprops
        if isinstance(tree,(SpudParser.PowwildwithpropsContext,SpudParser.BasicwildwithpropsContext)):
            return self._handle_name_and_props(tree.children[0],context)
        if isinstance(tree.children[0],antlr4.tree.Tree.TerminalNodeImpl):
            # can only be $
            name = None
        else:
            name = self._handle_all_names(tree.children[0],context)
        props = self._handle_all_props(tree.children[1],context)
        return name, props
    def _handle_lineingredient(self,tree,context):
        # tree is SpudParser.Lineingredient
        number = None
        unit = None
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                number = MyNumber.from_tree(child)
            elif isinstance(child,SpudParser.BasicnameContext):
                # units
                unit = self._handle_all_names(child,context)
            elif isinstance(child,SpudParser.BasicnamepluspropsContext):
                # name and props
                name, props_raw = self._handle_name_and_props(child,context)
                props = set()
                for prop in props_raw:
                    props.add(prop[0])
        if number is None:
            # unit is also None
            number = MyNumber.one
            unit = ["mealsworth"]
        elif unit is None:
            # unit is None
            # but number is defined
            unit = ["count"]
        if "multiply" in context:
            number = number * context["multiply"]
        self.ingredients.append( Ingredient(number,unit,name,props) )
    def _handle_rule_side_simple(self,tree,context):
        # tree is SpudParser.RuleleftsimpleContext
        # or SpudParser.RulerightsimpleContext
        # or SpudParser.Rulerightsimplewild
        # returns list of (name,props) tuples
        output = []
        for child in tree.children:
            if isinstance(child,(
                    SpudParser.PownamewithpropsContext,
                    SpudParser.BasicnamewithpropswildContext,
                    SpudParser.BasicwildwithpropsContext)):
                output.append(self._handle_name_and_props(child,context))
        return output
    def _handle_line_renaming_prefixing(self,tree,context,output_type):
        # tree is SpudParser.LinerenamingContext
        # or SpudParser.LineprefixingContext
        # output_type is either RenamingRuleOutput or PrefixingRuleOutput
        for child in tree.children:
            if isinstance(child,SpudParser.RuleleftsimpleContext):
                left = self._handle_rule_side_simple(child,context)
            elif isinstance(child,SpudParser.RulerightsimplewildContext):
                right = self._handle_rule_side_simple(child,context)
        for i in range(len(right)):
            ro = output_type(right[i][0])
            if len(right[i][1]) != 0:
                ro = PropertiesRuleOutput(ro,right[i][1])
            right[i] = ro
        for i in range(len(left)):
            pattern = SinglePattern(left[i][0],self._props_conversion(left[i][1]))
            self.rules.append(Rule(pattern,right))
    def _handle_lineinserting(self,tree,context):
        # tree is SpudParser.LineinsertingContext
        for child in tree.children:
            if isinstance(child,SpudParser.RuleleftsimpleContext):
                left = self._handle_rule_side_simple(child,context)
            elif isinstance(child,SpudParser.RulerightsimpleContext):
                right = self._handle_rule_side_simple(child,context)
        for i in range(len(left)):
            pattern = SinglePattern(left[i][0],self._props_conversion(left[i][1]))
            pattern_size = len(left[i][0])
            outputs = []
            for j in range(len(right)):
                ro = InsertingRuleOutput(pattern_size,right[j][0])
                if len(right[j][1]) != 0:
                    ro = PropertiesRuleOutput(ro,right[j][1])
                outputs.append(ro)
            self.rules.append(Rule(pattern,outputs))
    def _handle_singleconvertingleftelem(self,tree,context):
        # tree is SpudParser.SingleconvertingleftelemContext
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                number = MyNumber.from_tree(child)
            elif isinstance(child,SpudParser.PownamewithpropsContext):
                name, props = self._handle_name_and_props(child,context)
                return (number,name,props)
    def _handle_singleconvertingrightelem(self,tree,context):
        # tree is SpudParser.SingleconvertingrightelemContext
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                number = MyNumber.from_tree(child)
            elif isinstance(child,antlr4.tree.Tree.TerminalNodeImpl):
                if child.getText() == "$":
                    return None # ignore this otherwise
            elif isinstance(child,SpudParser.BasicwildwithpropsContext):
                name, props = self._handle_name_and_props(child,context)
                return (number,name,props)
    def _handle_singleconvertingleft(self,tree,context):
        # tree is SpudParser.SingleconvertingleftContext
        patterns = []
        for child in tree.children:
            if isinstance(child,SpudParser.SingleconvertingleftelemContext):
                patterns.append(self._handle_singleconvertingleftelem(child,context))
        return patterns
    def _handle_singleconvertingright(self,tree,context):
        # tree is SpudParser.SingleconvertingrightContext
        outputs = []
        for child in tree.children:
            if isinstance(child,SpudParser.SingleconvertingrightelemContext):
                outputs.append(self._handle_singleconvertingrightelem(child,context))
        return outputs
    def _handle_linesingleconverting(self,tree,context):
        # tree is SpudParser.LinesingleconvertingContext
        for child in tree.children:
            if isinstance(child,SpudParser.SingleconvertingleftContext):
                left = self._handle_singleconvertingleft(child,context)
            elif isinstance(child,SpudParser.SingleconvertingrightContext):
                right = self._handle_singleconvertingright(child,context)
        for i in range(len(left)):
            # left[i][0] is the number, not None
            # left[i][1] is the name
            # left[i][2] is the props
            denominator = left[i][0].multiplicative_inverse()
            pattern = SinglePattern(left[i][1],self._props_conversion(left[i][2]))
            outputs = []
            for j in range(len(right)):
                # elements of right are (number,name,props)
                # number may be None
                if right[j][0] is None:
                    ratio = MyNumber.one
                else:
                    ratio = denominator * right[j][0]
                ro = SingleConvertingRuleOutput(ratio,right[j][1])
                if len(right[j][2]) != 0:
                    ro = PropertiesRuleOutput(ro,right[j][2])
                outputs.append(ro)
            self.rules.append(Rule(pattern,outputs))
    def _handle_doubleconvertingleftsub_1_2(self,tree,context):
        # tree is SpudParser.Doubleconvertingleftsub1Context
        # or SpudParser.Doubleconvertingleftsub2Context
        for child in tree.children:
            if isinstance(child,antlr4.tree.Tree.TerminalNodeImpl):
                if child.getText() == "$":
                    unit = None
                    continue
            if isinstance(child,SpudParser.PownameContext):
                unit = self._handle_all_names(child,context)
            elif isinstance(child,
                (SpudParser.PownamewithpropsContext,
                SpudParser.PowwildwithpropsContext)):
                    name, props = self._handle_name_and_props(child,context)
                    return unit, name, props
    def _handle_doubleconvertingleftelem(self,tree,context):
        # tree is SpudParser.DoubleconvertingleftelemContext
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                number = MyNumber.from_tree(child)
            elif isinstance(child,
                (SpudParser.Doubleconvertingleftsub1Context,
                SpudParser.Doubleconvertingleftsub2Context)):
                    unit, name, props = self._handle_doubleconvertingleftsub_1_2(child,context)
                    return (number,unit,name,props)
    def _handle_doubleconvertingrightelem(self,tree,context):
        # tree is SpudParser.DoubleconvertingrightelemContext
        is_number_seen = False
        for child in tree.children:
            if isinstance(child,antlr4.tree.Tree.TerminalNodeImpl):
                if child.getText() == "$": # ignore otherwise
                    # but may be number or unit
                    # depending on context
                    if is_number_seen:
                        # it is the unit
                        unit = None
                    else:
                        # it is the number
                        is_number_seen = True
                        number = None
            elif isinstance(child,SpudParser.NumberContext):
                is_number_seen = True
                number = MyNumber.from_tree(child)
            elif isinstance(child,SpudParser.BasicnameContext):
                unit = self._handle_all_names(child,context)
            elif isinstance(child,SpudParser.BasicwildwithpropsContext):
                name, props = self._handle_name_and_props(child,context)
                return (number,unit,name,props)
    def _handle_doubleconvertingleft(self,tree,context):
        # tree is SpudParser.DoubleconvertingleftContext
        patterns = []
        for child in tree.children:
            if isinstance(child,SpudParser.DoubleconvertingleftelemContext):
                patterns.append(self._handle_doubleconvertingleftelem(child,context))
        return patterns
    def _handle_doubleconvertingright(self,tree,context):
        # tree is SpudParser.DoubleconvertingrightContext
        outputs = []
        for child in tree.children:
            if isinstance(child,SpudParser.DoubleconvertingrightelemContext):
                outputs.append(self._handle_doubleconvertingrightelem(child,context))
        return outputs
    def _handle_linedoubleconverting(self,tree,context):
        # tree is SpudParser.LinedoubleconvertingContext
        for child in tree.children:
            if isinstance(child,SpudParser.DoubleconvertingleftContext):
                left = self._handle_doubleconvertingleft(child,context)
            elif isinstance(child,SpudParser.DoubleconvertingrightContext):
                right = self._handle_doubleconvertingright(child,context)
        for i in range(len(left)):
            # left[i][0] is the number, not None
            # left[i][1] is the unit, may be None
            # left[i][2] is the name, may be None
            # left[i][3] is the props
            denominator = left[i][0].multiplicative_inverse()
            pattern = DoublePattern(left[i][1],left[i][2],self._props_conversion(left[i][3]))
            outputs = []
            for j in range(len(right)):
                if right[j][0] is None:
                    ratio = None
                else:
                    ratio = denominator * right[j][0]
                ro = DoubleConvertingRuleOutput(ratio,right[j][1],right[j][2])
                if len(right[j][3]) != 0:
                    ro = PropertiesRuleOutput(ro,right[j][3])
                outputs.append(ro)
            self.rules.append(Rule(pattern,outputs))
    def _handle_linestandardrule(self,tree,context):
        # tree is SpudParser.LinestandardruleContext
        for child in tree.children:
            if isinstance(child,SpudParser.LinerenamingContext):
                self._handle_line_renaming_prefixing(child,context,RenamingRuleOutput)
            elif isinstance(child,SpudParser.LineprefixingContext):
                self._handle_line_renaming_prefixing(child,context,PrefixingRuleOutput)
            elif isinstance(child,SpudParser.LineinsertingContext):
                self._handle_lineinserting(child,context)
            elif isinstance(child,SpudParser.LinesingleconvertingContext):
                self._handle_linesingleconverting(child,context)
            elif isinstance(child,SpudParser.LinedoubleconvertingContext):
                self._handle_linedoubleconverting(child,context)
    def _handle_atrun(self,tree,context):
        # tree is SpudParser.AtrunContext
        statement = tree.getText()
        # search and replace all @run with @import
        statement = "@import".join(statement.split("@run"))
        abs_path = self.resolve_import_path(context["BASE_DIR"],statement)
        self.recursive_load_parse(abs_path,True,context)
    def _handle_atsub1(self,tree,context):
        # tree is SpudPArser.Atsub1Context
        for child in tree.children:
            if isinstance(child,SpudParser.PownamewithpropsContext):
                name, props = self._handle_name_and_props(child,context)
                return DoublePattern(None,name,self._props_conversion(props))
    def _handle_atsub2(self,tree,context):
        # tree is SpudParser.Atsub2Context
        name = None
        props = {}
        for child in tree.children:
            if isinstance(child,SpudParser.PownameContext):
                unit = self._handle_all_names(child,context)
            elif isinstance(child,SpudParser.PowwildwithpropsContext):
                name, props = self._handle_name_and_props(child,context)
        return DoublePattern(unit,name,self._props_conversion(props))
    def _handle_atsub(self,tree,context):
        # tree is SpudParser.AtsubContext
        for child in tree.children:
            if isinstance(child,SpudParser.Atsub1Context):
                return self._handle_atsub1(child,context)
            if isinstance(child,SpudParser.Atsub2Context):
                return self._handle_atsub2(child,context)
    def _handle_athold(self,tree,context):
        # tree is SpudParser.AtholdContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtsubContext):
                pattern = self._handle_atsub(child,context)
                self.rules.append(HoldRule(pattern,0))
                # holds have priority 0
    def _handle_atholdunit(self,tree,context):
        # tree is SpudParser.AtholdunitContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtsubContext):
                pattern = self._handle_atsub(child,context)
                self.rules.append(HoldRule(pattern,4))
                # holdunits have priority 4
    def _handle_atdec(self,tree,context):
        # tree is SpudParser.AtdecContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtsubContext):
                pattern = self._handle_atsub(child,context)
                self.rules.append(Rule(pattern,[DecRuleOutputInstance]))
    def _handle_atfrac(self,tree,context):
        # tree is SpudParser.AtfracContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtsubContext):
                pattern = self._handle_atsub(child,context)
                self.rules.append(Rule(pattern,[FracRuleOutputInstance]))
    def _handle_atbeginmultiply(self,tree,context):
        # tree is SpudParser.AtbeginmultiplyContext
        for child in tree.children:
            if isinstance(child,SpudParser.NumberContext):
                value = MyNumber.from_tree(child)
                new_context = copy(context)
                if "multiply" in new_context:
                    value = new_context["multiply"] * value
                new_context["multiply"] = value
                return new_context
    def _handle_atbegin(self,tree,context):
        # tree is SpudParser.AtbeginContext
        # return a new context with updated values
        for child in tree.children:
            if isinstance(child,SpudParser.AtbeginisolateContext):
                # ignore all external forces
                return {"BASE_DIR":context["BASE_DIR"]}
            if isinstance(child,SpudParser.AtbeginmultiplyContext):
                return self._handle_atbeginmultiply(child,context)
        raise Exception("Internal Error")
    def _handle_atsection(self,tree,context):
        # tree is SpudParser.AtsectionContext
        for child in tree.children:
            # ignore atend
            if isinstance(child,SpudParser.AtbeginContext):
                context = self._handle_atbegin(child,context)
            elif isinstance(child,SpudParser.LineContext):
                self._handle_line(child,context)
    def _handle_lineat(self,tree,context):
        # tree is SpudParser.LineatContext
        for child in tree.children:
            if isinstance(child,SpudParser.AtrunContext):
                self._handle_atrun(child,context)
            elif isinstance(child,SpudParser.AtholdContext):
                self._handle_athold(child,context)
            elif isinstance(child,SpudParser.AtholdunitContext):
                self._handle_atholdunit(child,context)
            elif isinstance(child,SpudParser.AtdecContext):
                self._handle_atdec(child,context)
            elif isinstance(child,SpudParser.AtfracContext):
                self._handle_atfrac(child,context)
            elif isinstance(child,SpudParser.AtsectionContext):
                self._handle_atsection(child,context)
    def _handle_line(self,tree,context):
        # tree is a SpudParser.LineContext
        if tree.children is None:
            return
        for child in tree.children:
            if isinstance(child,SpudParser.LineingredientContext):
                self._handle_lineingredient(child,context)
            elif isinstance(child,SpudParser.LinestandardruleContext):
                self._handle_linestandardrule(child,context)
            elif isinstance(child,SpudParser.LineatContext):
                self._handle_lineat(child,context)
    def _handle_spud(self,tree,context):
        # tree is a SpudParser.SpudContext
        if tree.children is None:
            return
        for child in tree.children:
            if isinstance(child,SpudParser.LineContext):
                self._handle_line(child,context)
    def recursive_load_parse(self,abs_path,force=False,context=None):
        # abs_path is a string with the absolute file path to load
        tree = self.get_tree(abs_path,force)
        if tree is None:
            return
        imports = []
        self.extract_imports(imports,tree)
        if context is None:
            context = {}
        else:
            context = copy(context)
        context["BASE_DIR"] = os.path.dirname(abs_path)
        for statement in imports:
            self.recursive_load_parse(self.resolve_import_path(
                context["BASE_DIR"],
                statement
            ))
        self._handle_spud(tree,context)



# defined source_version_check function

def get_version_first_index(raw):
    # look for version~~ starting in the first 32 bytes
    # return first index of version if found
    # return None otherwise
    target = b'spudversion~~'
    for first_byte in range(32):
        match = True
        for i in range(len(target)):
            index = first_byte + i
            if index >= len(raw):
                return None
            if raw[index] != target[i]:
                match = False
                break
        if match:
            return first_byte + len(target)
    return None

def get_version_last_index(raw,version_index):
    # look for ~~ to end version
    # throw if hit end of input
    # or 128 byte maximum first
    tilde_count = 0
    # ~ is code 126
    for i in range(130):
        index = version_index + i
        if index >= len(raw):
            raise ValueError("Malformed input, expected ~~ but got end of input")
        if raw[index] == 126:
            tilde_count += 1
            if tilde_count == 2:
                # excluside final index
                return index - 1
        else:
            tilde_count = 0
    raise ValueError("Malformed input, expected ~~ but reached 128 byte limit")

def extract_version(raw):
    # look for version~~ starting in the first 32 bytes
    # return version as bytes if found
    # return None if no version found
    # return b'ERROR' if malformed
    first_index = get_version_first_index(raw)
    if first_index is None:
        return None
    last_index = get_version_last_index(raw,first_index)
    return raw[first_index:last_index]

this_verison = {
    "major": 0,
    "minor": 4,
    "patch": 1
}

def verify_version(version):
    global this_verison
    # compare
    sections = version.split(b'.')
    # we expect 1, 2, or 3 sections
    if len(sections) > 3:
        return False
    # they should all be nonegative integers
    for i in range(len(sections)):
        try:
            sections[i] = int(sections[i].decode("UTF8"))
        except:
            return False
        if sections[i] < 0:
            return False
    # they are all nonnegative integers now
    # pad for exactly 3 sections
    while len(sections) < 3:
        sections.append(0)
    # major version
    if sections[0] != this_verison["major"]:
        return False
    # major versions are the same
    if sections[1] < this_verison["minor"]:
        return True
    if sections[1] > this_verison["minor"]:
        return False
    # minor versions are the same
    if sections[2] < this_verison["patch"]:
        return True
    if sections[2] > this_verison["patch"]:
        return False
    # versions are exactly the same
    return True

def source_version_check(raw):
    # raw is a bytes object
    # corresponding to a recently loaded file
    # we want to know if the source file
    # belongs to a compatible version
    # no version indicated will be allowed
    # throws exception on malformed input
    # returns (True,) for compatible and (False,memo) for incompatible
    version = extract_version(raw)
    if version is None:
        return (True,)
    version_compat = verify_version(version)
    if version_compat:
        return (True,)
    else:
        global this_verison
        name = str(this_verison["major"]) + "." + str(this_verison["minor"]) + "." + str(this_verison["patch"])
        return (False,"This is Spudlang version: "+name+" The source reports a version of: "+version.decode("UTF8"))